/*
 * Copyright (c) 1999, Boris Popov
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Boris Popov.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: ipxping.c,v 1.1.1.1 1999/11/23 10:27:13 bp Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netipx/ipx.h>

#include <limits.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <err.h>
#include <errno.h>
#include <sysexits.h>

#ifndef IPXPORT_IPXPING
#define IPXPORT_IPXPING		0x9086
#endif

#define PING_DATA_LEN		8


static char pingdata[PING_DATA_LEN] = "FreeBSD ";

static int ipx_recv(int sock, void *buf, int len, int flags, int timeout);
static void usage(void);


int
main(int argc, char *argv[]) {
	int opt, sock, slen, rlen, count, delay, nreceived, nsent, i;
	u_short seq;
	u_char rq[10 + PING_DATA_LEN], rp[10 + PING_DATA_LEN], *p;
	char *ep;
	struct ipx ipx;
	struct ipx_addr dest;
	struct sockaddr_ipx sipx;
	struct timeval now, last;

	delay = 1;
	count = -1;
	seq = 0;

	while ((opt = getopt(argc, argv, "c:i:")) != EOF) {
		switch (opt) {
		    case 'c':
			i = strtoul(optarg, &ep, 0);
			if (*ep || ep == optarg || i > LONG_MAX || !i)
				errx(EX_USAGE,
				    "invalid count of packets to transmit: `%s'",
				    optarg);
			count = i;
			break;
		    case 'i':
			i = strtoul(optarg, &ep, 0);
			if (*ep || ep == optarg || i > LONG_MAX || !i)
				errx(EX_USAGE,
				    "invalid delay interval: `%s'",
				    optarg);
			delay = i;
			break;
		    default:
			usage();
			/* NOTREACHED */
		}
	}

	if (argc - optind != 1)
		usage();
	dest = ipx_addr(argv[optind]);
	if (ipx_nullnet(dest) || ipx_nullhost(dest))
		errx(EX_DATAERR, "invalid address");

	sock = socket(AF_IPX, SOCK_DGRAM, 0);
	if (sock < 0)
		return (errno);
	slen = sizeof(sipx);
	bzero(&sipx, slen);
	sipx.sipx_family = AF_IPX;
	sipx.sipx_len = slen;
	if (bind(sock, (struct sockaddr*)&sipx, slen) == -1) {
		close(sock);
		return(errno);
	}

	ipx.ipx_pt = IPXPROTO_PXP;
	if (setsockopt(sock, 0, SO_DEFAULT_HEADERS, &ipx, sizeof(ipx)))
		errx(0, "setsockopt()");

	bzero(&sipx, slen);
	sipx.sipx_family = AF_IPX;
	sipx.sipx_len = slen;
	sipx.sipx_addr = dest;
	sipx.sipx_addr.x_port = htons(IPXPORT_IPXPING);
	printf("Ping %s\n\n", ipx_ntoa(sipx.sipx_addr));

	bcopy("Ping", rq, 4);
	p = rq + 4;
	*p++ = 1;	/* version */
	*p++ = 0;	/* request */
	p += 2;
	*p++ = 0;	/* result */
	*p++ = 0;	/* reserved */
	bcopy(pingdata, p, PING_DATA_LEN);
	nreceived = 0;
	nsent = 0;
	do {
		*(u_short*)(rq + 6) = seq;
		(void)gettimeofday(&last, NULL);
		rlen = sendto(sock, rq, sizeof(rq), 0,
			(struct sockaddr*)&sipx, slen);
		if (rlen < 0) {
			warn("Can't sendto");
		} else
			nsent++;

		rlen = ipx_recv(sock, rp, sizeof(rp), 0, delay);
		if (rlen == -1) {
			warn("can't recv()");
		} else if (rlen < sizeof(rp)) {
			warn("reply too short");
		} else if (memcmp("Ping", rp , 4) != 0) {
			warnx("signature mismatch");
		} else if (memcmp(pingdata, rp + 10, PING_DATA_LEN) != 0) {
			warnx("reply data mismatch");
		} else if (*(u_short*)(rp + 6) != seq) {
			warnx("reply sequence mismatch: %d:%d", 
			    *(u_short*)(rp + 6), seq);
		} else {
			nreceived++;
			printf("%d bytes received\n", rlen);
		}

		if (count != -1) {
			count--;
			if (count == 0)
				break;
		}
		seq++;
		(void)gettimeofday(&now, NULL);
		i = delay - (now.tv_sec - last.tv_sec);
		if (i > 0)
			sleep(i);
	} while(1);

	printf("\n%d packets sent, %d received\n", nsent, nreceived);
	exit(nreceived ? 0 : 1);
}

static int
ipx_recv(int sock, void *buf, int len, int flags, int timeout) {
	fd_set rd, wr, ex;
	struct timeval tv;
	int result;

	FD_ZERO(&rd);
	FD_ZERO(&wr);
	FD_ZERO(&ex);
	FD_SET(sock, &rd);

	tv.tv_sec = timeout;
	tv.tv_usec = 0;

	if ((result = select(sock + 1, &rd, &wr, &ex, &tv)) == -1) {
		return result;
	}
	if (FD_ISSET(sock, &rd)) {
		result = recv(sock, buf, len, flags);
	} else {
		errno = ETIMEDOUT;
		result = -1;
	}
	return result;
}

static void
usage(void) {
	fprintf(stderr, "%s\n",
	    "usage: ipxping [-c count] [-i delay] network.host"
	);
	exit(EX_USAGE);
}
